#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A add_column step, which allows you to add a column to 
a dataframe. 
"""

from mitosheet.utils import create_new_step, does_sheet_index_exist_within_step
from mitosheet.errors import (
    make_column_exists_error,
    make_no_sheet_error
)

ADD_COLUMN_DISPLAY_NAME = 'Added a Column'

ADD_COLUMN_EVENT = 'add_column_edit'
ADD_COLUMN_STEP_TYPE = 'add_column'

ADD_COLUMN_PARAMS = [
    'sheet_index', # int
    'column_header', # the new column to create
    'column_header_index' # the index of the new column
]

def execute_add_column_step(
        curr_step,
        sheet_index,
        column_header,
        column_header_index
    ):
    """
    The function responsible for updating the widget state container
    with a new add_column step.
    """
    # if the sheet doesn't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    if column_header in curr_step['column_metatype'][sheet_index]:
        raise make_column_exists_error(column_header)

    # We add a new step with the added column
    new_step = create_new_step(curr_step, ADD_COLUMN_STEP_TYPE)

    # If the column_header_index is out of range, then make the new column the last column
    if column_header_index < 0 or len(curr_step['dfs'][sheet_index].columns) <= column_header_index:
        column_header_index = len(curr_step['dfs'][sheet_index].columns)

    # Update the state variables
    new_step['column_metatype'][sheet_index][column_header] = 'formula'
    new_step['column_type'][sheet_index][column_header] = 'number'
    new_step['column_spreadsheet_code'][sheet_index][column_header] = '=0'
    new_step['column_python_code'][sheet_index][column_header] = f'df[\'{column_header}\'] = 0'
    new_step['column_evaluation_graph'][sheet_index][column_header] = set()
    new_step['column_filters'][sheet_index][column_header] = {'operator': 'And', 'filters': []}
        
    # Update the dataframe
    new_step["dfs"][sheet_index].insert(column_header_index, column_header, 0)
    
    return new_step


def transpile_add_column_step(
        step,
        sheet_index,
        column_header, 
        column_header_index
    ):
    """
    Transpiles an add column step to python code!
    """
    # If the column_header_index is out of range, then make the new column the last column
    # Note: here we subtract 1 from the len of the columns arrray because the array generated from 
    # the step dfs object already inclused the column header that the transpiled code is adding.
    if column_header_index < 0 or len(step['dfs'][sheet_index].columns) <= column_header_index:
        column_header_index = len(step['dfs'][sheet_index].columns) - 1

    return [
        f'{step["df_names"][sheet_index]}.insert({column_header_index}, \'{column_header}\', 0)'
    ]


def describe_add_column_step(
        sheet_index,
        column_header,
        column_header_index,
        df_names=None
    ):    
    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Added column {column_header} to {df_name}'
    return f'Added column {column_header}'


ADD_COLUMN_STEP = {
    'step_version': 2,
    'step_display_name': ADD_COLUMN_DISPLAY_NAME,
    'event_type': ADD_COLUMN_EVENT,
    'step_type': ADD_COLUMN_STEP_TYPE,
    'params': ADD_COLUMN_PARAMS,
    'saturate': None,
    'execute': execute_add_column_step,
    'transpile': transpile_add_column_step,
    'describe': describe_add_column_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A undo update, which removes the most recent step from
the sheet. 
"""

UNDO_EVENT = 'undo'
UNDO_PARAMS = []

def execute_undo_update(wsc):
    """
    The function responsible for updating the widget state container
    by removing the most recent step.

    If there is no most recent step, does nothing.
    """

    if len(wsc.steps) == 1:
        return 

    # Pop off the last element
    wsc.steps.pop()

    # TODO: there is currently a bug with this event when we are rolled back, this will
    # jump us to the end of the steps
    wsc.curr_step_idx = len(wsc.steps) - 1


"""
This object wraps all the information
that is needed for a undo step!
"""
UNDO_UPDATE = {
    'event_type': UNDO_EVENT,
    'params': UNDO_PARAMS,
    'execute': execute_undo_update
}








#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A rename dataframe step changes the name of a specific dataframe
at a specific index.
"""

from mitosheet.utils import create_new_step, make_valid_header
from mitosheet.steps.dataframe_steps.dataframe_duplicate import get_first_unused_name

DATAFRAME_RENAME_DISPLAY_NAME = 'Renamed a Dataframe'

DATAFRAME_RENAME_EVENT = 'dataframe_rename_edit'
DATAFRAME_RENAME_STEP_TYPE = 'dataframe_rename'

DATAFRAME_RENAME_PARAMS = [
    'sheet_index',
    'old_dataframe_name', # this is saturated 
    'new_dataframe_name'
]

def get_valid_python_variable_name(suggested_dataframe_name):
    """
    Turns the suggested dataframe name into a valid Python identifier.
    """
    # See: https://www.w3schools.com/python/ref_string_isidentifier.asp
    if suggested_dataframe_name.isidentifier():
        return suggested_dataframe_name

    # First, we try and just fill in the non valid characters (like a column header)
    valid_header_version = make_valid_header(suggested_dataframe_name)
    if valid_header_version.isidentifier():
        return valid_header_version

    # Finially prefix with df, so that it doesn't start with a number (as vars cannot)
    return f'df_{valid_header_version}'


def get_valid_dataframe_name(df_names, suggested_dataframe_name):
    """
    Given an attempt at a dataframe name, makes sure that is a valid
    dataframe name (e.g. it is a valid Python identifier, and it is
    not a duplicate).
    """
    valid_identifier = get_valid_python_variable_name(suggested_dataframe_name)
    return get_first_unused_name(df_names, valid_identifier)

def saturate_dataframe_rename(
        curr_step,
        event
    ):
    """
    Saturates the dataframe_rename_edit, by adding the old_dataframe_name
    to the event.
    """
    sheet_index = event['sheet_index']
    old_dataframe_name = curr_step['df_names'][sheet_index]
    event['old_dataframe_name'] = old_dataframe_name


def execute_dataframe_rename(
        curr_step,
        sheet_index,
        old_dataframe_name,
        new_dataframe_name
    ):
    """
    Actually performs the exection by creating a new step, and performing the
    dataframe rename in this new step.

    NOTE: this step does it best not to fail and cause annoying errors, 
    as annoying errors are annoying. So, if the given new_dataframe_name is
    invalid in any way, it will correct it to a valid dataframe name. 
    """

    # Bail early, if there is no change
    if old_dataframe_name == new_dataframe_name:
        return

    # Create a new step and save the parameters
    new_step = create_new_step(curr_step, DATAFRAME_RENAME_STEP_TYPE, deep=False)

    # Save the parameters
    new_step['sheet_index'] = sheet_index
    new_step['old_dataframe_name'] = old_dataframe_name
    new_step['new_dataframe_name'] = new_dataframe_name

    new_step['df_names'][sheet_index] = get_valid_dataframe_name(new_step['df_names'], new_dataframe_name)

    return new_step

def transpile_dataframe_rename(
        step,
        sheet_index,
        old_dataframe_name,
        new_dataframe_name
    ):
    """
    Transpiles as dataframe_rename step
    """

    return [f'{step["df_names"][sheet_index]} = {old_dataframe_name}']


def describe_dataframe_rename(
        sheet_index,
        old_dataframe_name,
        new_dataframe_name,
        df_names=None
    ):

    return f'Renamed {old_dataframe_name} to {new_dataframe_name}'


DATAFRAME_RENAME_STEP = {
    'step_version': 1,
    'step_display_name': DATAFRAME_RENAME_DISPLAY_NAME,
    'event_type': DATAFRAME_RENAME_EVENT,
    'step_type': DATAFRAME_RENAME_STEP_TYPE,
    'params': DATAFRAME_RENAME_PARAMS,
    'saturate': saturate_dataframe_rename,
    'execute': execute_dataframe_rename,
    'transpile': transpile_dataframe_rename,
    'describe': describe_dataframe_rename
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions
"""
from copy import deepcopy
import json
import pandas as pd
import numpy as np
import numbers
from string import ascii_letters, digits
import subprocess
import os
import uuid
import re

def make_valid_header(column_header):
    """
    Takes a header, and performs replaces against common characters
    to make the column_header valid!
    """
    # If it's just numbers, turn it into a string (with an underscore)
    if isinstance(column_header, numbers.Number):
        return str(column_header).replace('.', '_') + '_'

    # If it's just numbers in a string, add an underscore
    if column_header.isdigit():
        return column_header + "_"

    # Replace common invalid seperators with valid seperators
    replace_dict = {
        ' ': '_',
        '-': '_',
        '(': '_',
        ')': '_',
        '/': '_',
        '#': 'num',
        ',': '_',
        '.': '_',
        '!': '_',
        '?': '_'
    }
    for find, replace in replace_dict.items():
        column_header = column_header.replace(find, replace)
    
    if not is_valid_header(column_header):
        # Because we detect column headers using a word match, any word character counts
        # in a valid character
        pattern = re.compile("\w")

        new_header = ''.join([
            c for c in column_header if pattern.search(c)
        ])
        if not is_valid_header(new_header):
            # And then append an underscore, for good measure, and this should fix it!
            new_header = new_header + '_'

        return new_header
    return column_header

def is_valid_header(column_header):
    """
    A header is valid if It is a string that is made up of all word characters,
    with at least one non numeric char, and has at least one char.

    Valid examples: A, ABC, 012B, 213_bac, 123_123
    Invalid examples: 123, 123!!!, ABC!, 123-123

    This is a result of not being able to distingush column headers from constants
    otherwise, and would not be necessary if we had a column identifier!
    """
    
    # Note the start and end characters in the regex, to make sure it's a full match
    return isinstance(column_header, str) and \
        len(column_header) > 0 and \
        re.compile("^\w+$").search(column_header) and \
        not column_header.isdigit()


def get_invalid_headers(df: pd.DataFrame):
    """
    Given a dataframe, returns a list of all the invalid headers this list has. 
    """
    return [
        header for header in df.columns.tolist()
        if not is_valid_header(header)
    ]

def get_header_renames(column_headers):
    """
    Given a list of column headers, returns a mapping from old, invalid headers to
    new, valid headers. Empty if no renames are necessary.
    """
    renames = dict()
    for column_header in column_headers:
        if not is_valid_header(column_header):
            valid_header = make_valid_header(column_header)
            renames[column_header] = valid_header

    return renames


def dfs_to_json(dfs):
    return json.dumps([df_to_json_dumpsable(df) for df in dfs])


def df_to_json_dumpsable(df: pd.DataFrame):
    """
    Returns a dataframe represented in a way that can be turned into a 
    JSON object with json.dumps
    """
    # First, we get the head, and then we make a copy (as we modify the df below)
    # NOTE: we get the head first for efficiency reasons.
    df = df.head(n=2000).copy(deep=True) # we only show the first 2k rows!
    
    # Second, we figure out which of the columns contain dates, and we
    # convert them to string columns (for formatting reasons).
    # NOTE: we don't use date_format='iso' in df.to_json call as it appends seconds to the object, 
    # see here: https://stackoverflow.com/questions/52730953/pandas-to-json-output-date-format-in-specific-form
    date_columns = df.select_dtypes(include=['datetime64'])
    for key in date_columns:
        df[key] = date_columns[key].dt.strftime('%Y-%m-%d %X')

    # Third, we figure out which of the columns contain timedeltas, and 
    # we format the timedeltas as strings to make them readable
    date_diff_columns = df.select_dtypes(include=['timedelta64[ns]'])
    for key in date_diff_columns:
        df[key] = date_diff_columns[key].apply(lambda x: str(x))

    json_obj = json.loads(df.to_json(orient="split"))
    # Then, we go through and find all the null values (which are infinities),
    # and set them to 'NaN' for display in the frontend.
    for d in json_obj['data']:
        for idx, e in enumerate(d):
            if e is None:
                d[idx] = 'NaN'

    return json_obj


def get_column_filter_type(series):
    """
    Given a series, take's the series dtype and catergorizes the correct
    column_filter_type so that the filter modal can display only
    valid filtering options
    
    TODO: unify this with the other Mito types!
    """
    try: 
        if series.dtype == np.float64 or series.dtype == np.int64:
            return 'number'
        elif series.dtype == 'datetime64[ns]':
            return 'datetime'
        elif series.dtype == np.object:
            return 'string'
    except:
        # If there is an error, we just call this a string column. In the wild, we have
        # seen errors where dtype is not defined on the df object, but we have been
        # unable to replicate this or figure out when it occurs.
        return 'string'
    
    # TODO: what do we do if it's none? for now just return string, for no reason.
    return 'string'

def get_app_directory_from_jupyter_lab_path_output(output):
    """
    Takes the string output of `jupyter lab path` and returns the application folder for jupyter lab.

    The output looks like:
    Application directory:   <application path (depends on what system you are on)>
    User Settings directory: <user settings path (depends on what system you are on)>
    Workspaces directory: <workspace directory path (depends on what system you are on)>
    
    NOTE: this _should_ be able to handle any path we might want, including spaces, or Linux
    or Windows or Mac paths.
    """
    return [x for x in output.split('\n') if x.startswith('Application directory:')][0][len('Application directory:'):].strip()


def is_imported_correctly():
    """
    Helper function to determine if Mito is imported correctly. For this to be the case, 
    Mito must be installed and enabled on a JupyterLab 2.0 instance, along with 
    @jupyter-widgets/jupyterlab-manager.

    We check this by finding out where jupyter lab is installed, and then seeing what 
    extensions are installed. This is not perfect, but should detect most common cases
    """
    try:
        # First, we get the directory that Jupyter is installed in
        result = subprocess.run(["jupyter", "lab", "path"], capture_output=True)
        stdout = result.stdout.decode()
        app_directory = get_app_directory_from_jupyter_lab_path_output(stdout)
        extension_directory = os.path.join(app_directory, 'extensions')
        # Then, we check if the correct extensions are in the extension directory
        if os.path.exists(extension_directory):
            extensions = os.listdir(extension_directory)
            includes_mito = len([extension for extension in extensions if extension.startswith('mitosheet')]) > 0
            includes_jupyter_widgets = len([extension for extension in extensions if extension.startswith('jupyter-widgets-jupyterlab-manager')]) > 0
            return includes_mito and includes_jupyter_widgets
        
        return False
    except:
        return False
    

def does_sheet_index_exist_within_step(step, sheet_index):
    """
    Returns true iff a sheet_index exists within a step
    """
    return not (sheet_index < 0 or sheet_index >= len(step['dfs']))


def add_df_to_step(step, new_df, df_name=None):
    """
    Helper function for adding a new dataframe to the current step!
    """
    # Update dfs by appending new df
    step['dfs'].append(new_df)
    # Also update the dataframe name
    if df_name is None:
        step['df_names'].append(f'df{len(step["df_names"]) + 1}')
    else:
        step['df_names'].append(df_name)
    
    # Update all the variables that depend on column_headers
    column_headers = new_df.keys()
    step['column_metatype'].append({column_header: 'value' for column_header in column_headers})
    step['column_type'].append({column_header: get_column_filter_type(new_df[column_header]) for column_header in column_headers})
    step['column_spreadsheet_code'].append({column_header: '' for column_header in column_headers})
    step['column_python_code'].append({column_header: '' for column_header in column_headers})
    step['column_evaluation_graph'].append({column_header: set() for column_header in column_headers})
    step['column_filters'].append({column_header: {'operator':'And', 'filters': []} for column_header in column_headers})

def create_new_step(prev_step, step_type, deep=True):
    """
    Creates a new step with new_step_id and step_type that starts
    with the ending state of the previous step

    We default to deep=True, which makes a deep copy of the dataframes
    that are in the previous step. However, for steps that don't modify
    any existing dataframes (e.g. imports, merges, pivots), we can 
    set this to deep=False/
    """
    # The new step is a copy of the previous step, where we only take the data we need
    # (which is the formula content only)
    new_step = dict()

    new_step['step_type'] = step_type
    new_step['column_metatype'] = deepcopy(prev_step['column_metatype'])
    new_step['column_type'] = deepcopy(prev_step['column_type'])
    new_step['column_spreadsheet_code'] = deepcopy(prev_step['column_spreadsheet_code'])
    new_step['added_columns'] = [[] for df in prev_step['dfs']]
    new_step['column_python_code'] = deepcopy(prev_step['column_python_code'])
    new_step['column_evaluation_graph'] = deepcopy(prev_step['column_evaluation_graph'])
    new_step['column_filters'] = deepcopy(prev_step['column_filters'])
    new_step['dfs'] = [df.copy(deep=deep) for df in prev_step['dfs']]
    new_step['df_names'] = deepcopy(prev_step['df_names'])

    return new_step


def get_random_id():
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def get_new_step_id():
    return str(uuid.uuid4())





import io
import base64
import pandas as pd
from mitosheet.sheet_functions.types.utils import get_mito_type
import matplotlib.pyplot as plt


def get_column_summary_graph(send, event, wsc):
    """
    Sends back a string representation of a PNG graph of the column_header 
    in the df at sheet_index in the widget state container.

    If the series is a numeric series, will return a histogram. Otherwise, as
    long as there are less than 20 distinct items in the series, will return
    a bar chart of the value count. Otherwise, will return nothing.
    """
    # We turn off interactive plotting for this function, to avoid graphs 
    # being printed to the console. However we make sure to turn it back on 
    # at the end if it was originally on, to not mess with any users code 
    # in the rest of their notebook
    was_interactive = plt.isinteractive()
    if was_interactive:
        plt.ioff()

    sheet_index = event['sheet_index']
    column_header = event['column_header']
    
    series: pd.Series = wsc.dfs[sheet_index][column_header]

    try:
        mito_type = get_mito_type(series)

        if mito_type == 'number_series':
            ax = series.plot.hist()
        else:
            if series.nunique() > 20:
                # Send an empty response
                send({
                    'event': 'api_response',
                    'id': event['id'],
                    'data': ''
                })
                # Then return so we only respond with an empty string
                return
            ax = series.value_counts().plot.bar()

        ax.set_title(column_header + ' frequencies')

        # From here: https://stackoverflow.com/questions/38061267/matplotlib-graphic-image-to-base64
        pic_IObytes = io.BytesIO()
        ax.figure.savefig(pic_IObytes, format='png', transparent=True, bbox_inches='tight')
        pic_IObytes.seek(0)
        pic_hash = base64.b64encode(pic_IObytes.read()).decode("utf-8").replace("\n", "")
        # Close all graphs so they don't each other write each
        plt.close('all')
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': pic_hash
        })
    except Exception as e:
        # As not being able to make a graph is a non-critical error that doesn't
        # result from user interaction, we don't want to throw an error if something
        # weird happens, so we just return nothing in this case
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': ''
        })
    
    # Turn interactive plotting back on, if it was on in the first place,
    # so we dont mess with users matplotlib code
    if was_interactive:
        plt.ion()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains a list of all the saved analyses that we store in code, for usage 
in the great big Mito machine. 

Just kidding. It's so we can easily distribute them without a ton of extra 
work and having to worry about eventually changing the .mito file. 
"""

from mitosheet.saves.mito_simple_raw_import import MITO_SIMPLE_RAW_IMPORT_SAVE
from mitosheet.saves.tutorial_v3.mito_tutorial_checkpoint_2 import MITO_TUTORIAL_CHECKPOINT_2
from mitosheet.saves.tutorial_v3.mito_tutorial_checkpoint_3 import MITO_TUTORIAL_CHECKPOINT_3
from mitosheet.saves.tutorial_v3.mito_tutorial_checkpoint_4 import MITO_TUTORIAL_CHECKPOINT_4
from mitosheet.saves.tutorial_v3.mito_tutorial_checkpoint_5 import MITO_TUTORIAL_CHECKPOINT_5
from mitosheet.saves.tutorial_v3.mito_tutorial_checkpoint_6 import MITO_TUTORIAL_CHECKPOINT_6

SAVES = {
    MITO_SIMPLE_RAW_IMPORT_SAVE['name']: MITO_SIMPLE_RAW_IMPORT_SAVE['saved_analysis'],
    MITO_TUTORIAL_CHECKPOINT_2['name']: MITO_TUTORIAL_CHECKPOINT_2['saved_analysis'],
    MITO_TUTORIAL_CHECKPOINT_3['name']: MITO_TUTORIAL_CHECKPOINT_3['saved_analysis'],
    MITO_TUTORIAL_CHECKPOINT_4['name']: MITO_TUTORIAL_CHECKPOINT_4['saved_analysis'],
    MITO_TUTORIAL_CHECKPOINT_5['name']: MITO_TUTORIAL_CHECKPOINT_5['saved_analysis'],
    MITO_TUTORIAL_CHECKPOINT_6['name']: MITO_TUTORIAL_CHECKPOINT_6['saved_analysis']
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains helpful utility functions for getting and setting the user profile 
in user.json
"""

import os
import json

# Where all global .mito files are stored
MITO_FOLDER = os.path.expanduser("~/.mito")

# The path of the user.json file
USER_JSON_PATH = os.path.join(MITO_FOLDER, 'user.json')

# The path of the user_install.json, which notably exists if the
# user has used the installer
USER_INSTALL_PATH = os.path.join(MITO_FOLDER, 'user_install.json')


def get_user_field(field, path=USER_JSON_PATH):
    """
    Returns the value stored at field in the user.json file,
    but may read a different file if it passed
    """
    try:
        with open(path) as f:
            return json.load(f)[field]
    except: 
        return None

def set_user_field(field, value):
    """
    Updates the value of a specific feild in user.json
    """
    with open(USER_JSON_PATH, 'r') as user_file_old:
        old_user_json = json.load(user_file_old)
        old_user_json[field] = value
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(old_user_json))




def get_column_dtype(send, event, wsc):
    """
    Sends back the dtype of the column
    """
    sheet_index = event['sheet_index']
    column_header = event['column_header']

    series = wsc.dfs[sheet_index][column_header]

    send({
        'event': 'api_response',
        'id': event['id'],
        'data': str(series.dtype)
    })



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains interfaces and helpers for dealing with the user profile.

Currently, the users profile is stored on disk where Mito is installed, 
but in the future we will support other profile storing locations (e.g.
on a server, so that users can log into a webapp!)
"""
import getpass
import os
import json
from datetime import datetime

from mitosheet.utils import get_random_id
from mitosheet.mito_analytics import log

from mitosheet._version import __version__
from mitosheet.user.user_utils import (
    USER_INSTALL_PATH, get_user_field, set_user_field,
    MITO_FOLDER, USER_JSON_PATH
)

# This is the default user json object
USER_JSON_DEFAULT = {
    'user_json_version': 1,
    'static_user_id': '',
    # A random secret that the user can use as salt when hashing things
    'user_salt': get_random_id(),
    'user_email': '',
    # A list of actions the user intends to do on the tool, which they fill
    # out when they sign up
    'intended_behavior': [],
    'received_tours': [],
    # A list of all the feedback the user has given
    'feedbacks': [],
    # If the user opted out of feedback, we store that they opted out, so
    # that we don't bombard them with feedback
    'closed_feedback': False,
    'mitosheet_current_version': __version__,
    'mitosheet_last_upgraded_date': datetime.today().strftime('%Y-%m-%d'),
    'mitosheet_last_five_usages': [datetime.today().strftime('%Y-%m-%d')]
}


def is_on_kuberentes_mito():
    """
    Returns True if the user is on Kuberentes Mito
    """
    user = getpass.getuser()
    return user == 'jovyan'


def is_local_deployment():
    """
    Helper function for figuring out if this a local deployment or a
    Mito server deployment
    """
    return not is_on_kuberentes_mito()  


def initialize_user():
    """
    Internal helper function that gets called every time mitosheet 
    is imported.

    It:
    1. Creates a ~/.mito folder, if it does not exist.

    2. Creates new, default user.json file if it does not exist, taking
       special care to do things properly if we're in a CI enviornment.

       Notably, if the user has used the installer, it sets the user
       id to be the same as the installed id, so that we can understand
       how users make it through installation

    3. Updates the user.json file with any new variables, as well as logging
       this usage, and any potential upgrades that may have occured
    """
    if not os.path.exists(MITO_FOLDER):
        os.mkdir(MITO_FOLDER)

    # Read in the static user id generated by the installer, which is None
    # if the user did not use the installer
    static_user_id_install = get_user_field('static_user_id_install', USER_INSTALL_PATH)

    # We only create a user.json file if it does not exist
    if not os.path.exists(USER_JSON_PATH):
        # First, we write an empty default object
        with open(USER_JSON_PATH, 'w+') as f:
            f.write(json.dumps(USER_JSON_DEFAULT))

        # Then, we create a new static id and capture the email for the user. 
        # We take special care to put all the CI enviornments 
        # (e.g. Github actions) under one ID and email
        if 'CI' in os.environ and os.environ['CI'] is not None:
            static_user_id = 'github_action'
            user_email = 'github@action.com'
        else:
            # Take the static user id from the installer, if it exists, and otherwise
            # generate a new one
            static_user_id = get_random_id() if static_user_id_install is None else static_user_id_install
            # We used to read the user email if they were signed in on a kubernetes
            # cluster, but instead we ask the user to go through the full signup flow
            # to make sure they accept the privacy policy, and get appropraite tours
            user_email = ''

        set_user_field('static_user_id', static_user_id)
        set_user_field('user_email', user_email)


    # Then we just make sure that the user.json has all the fields it needs defined
    # and if they are not defined, it sets them to the default values
    for field, default_value in USER_JSON_DEFAULT.items():
        if get_user_field(field) is None:
            set_user_field(field, default_value)

    # Then, we check if Mito has been upgraded since it was last imported
    # and if it has been upgraded, we upgrade the version and the upgrade date
    mitosheet_current_version = get_user_field('mitosheet_current_version')
    if mitosheet_current_version != __version__:
        set_user_field('mitosheet_current_version', __version__)
        set_user_field('mitosheet_last_upgraded_date', datetime.today().strftime('%Y-%m-%d'))
        # Log the upgrade. Note that this runs when the user _actually_ changes
        # the version of mitosheet that they are using, not just when they 
        # click the upgrade button in the app (although clicking this upgrade
        # button will stop the upgrade popup from showing up)
        log('upgraded_mitosheet', {'old_version': mitosheet_current_version, 'new_version': __version__})

    # We also note this import as a Mito usage, making sure to only 
    # mark this as usage once per day
    last_five_usages = get_user_field('mitosheet_last_five_usages')
    if len(last_five_usages) == 0 or last_five_usages[-1] != datetime.today().strftime('%Y-%m-%d'):
        last_five_usages.append(datetime.today().strftime('%Y-%m-%d'))
    # Then, we take the 5 most recent (or as many as there are), and save them
    if len(last_five_usages) < 5:
        most_recent_five = last_five_usages
    else: 
        most_recent_five = last_five_usages[-5:]
    set_user_field('mitosheet_last_five_usages', most_recent_five)


def should_upgrade_mitosheet():
    """
    A helper function that calculates if a user should upgrade,
    which in practice does this with the following heuristic:
    1. If the user has not upgraded in two weeks, then we always reccomend that the user
       upgrades.
    2. If the user has used the tool 4 times since they last upgraded, then we also reccomend
       that they upgrade

    NOTE: this should always return false if it is not a local installation, for obvious
    reasons.

    NOTE: if the user clicks the upgrade button in the app, then we change the upgraded 
    date to this date, so that the user doesn't get a bunch of annoying popups. This just
    pushes back when they are annoyed to upgrade!

    The motivation here is just: we want them to upgrade frequently, but we also don't
    want to just bombard them with upgrade messages. This is a nice middle ground.
    """
    if not is_local_deployment():
        return False

    mitosheet_last_upgraded_date = datetime.strptime(get_user_field('mitosheet_last_upgraded_date'), '%Y-%m-%d')
    mitosheet_last_five_usages = [datetime.strptime(usage, '%Y-%m-%d') for usage in get_user_field('mitosheet_last_five_usages')]

    current_time = datetime.now()
    # Condition (1)
    if (current_time - mitosheet_last_upgraded_date).days > 14:
        return True
    # Condition (2)
    elif len(mitosheet_last_five_usages) >= 4:
        # As this list is chronological, we just need to check 4 back
        if mitosheet_last_five_usages[-4] > mitosheet_last_upgraded_date:
            return True

    return False


def should_display_feedback():
    """
    Returns true if the user has used the tool on 3 seperate
    days and has not yet closed the feedback taskpane.
    """
    num_feedbacks = len(get_user_field('feedbacks'))
    num_days = len(get_user_field('mitosheet_last_five_usages'))
    closed_feedback = get_user_field('closed_feedback')

    return num_feedbacks <= 1 and num_days == 3 and not closed_feedback




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
This steps duplicates a dataframe of a given index. 
"""

from mitosheet.utils import add_df_to_step, create_new_step


DATAFRAME_DUPLICATE_DISPLAY_NAME = 'Duplicated a Dataframe'

DATAFRAME_DUPLICATE_EVENT = 'dataframe_duplicate_edit'
DATAFRAME_DUPLICATE_STEP_TYPE = 'dataframe_duplicate'

DATAFRAME_DUPLICATE_PARAMS = [
    'sheet_index'
]

def get_first_unused_name(df_names, dataframe_name):
    """
    Appends _1, _2, .. to df name until it finds an unused 
    dataframe name. If no append is necessary, will just return
    """
    if dataframe_name not in df_names:
        return dataframe_name

    for i in range(len(df_names) + 1):
        new_name = f'{dataframe_name}_{i + 1}'
        if new_name not in df_names:
            return new_name


def execute_dataframe_duplicate(
        curr_step,
        sheet_index
    ):
    """
    Duplicates the dataframe at sheet_index.
    """
    # Create a new step and save the parameters
    new_step = create_new_step(curr_step, DATAFRAME_DUPLICATE_STEP_TYPE, deep=False)

    # Save the parameters
    new_step['sheet_index'] = sheet_index

    # Execute the step
    df_copy = new_step['dfs'][sheet_index].copy(deep=True)
    new_name = get_first_unused_name(new_step['df_names'], new_step['df_names'][sheet_index] + '_copy')
    add_df_to_step(new_step, df_copy, new_name)

    return new_step


def transpile_dataframe_duplicate(
        step,
        sheet_index
    ):
    """
    Transpiles the dataframe duplication to Python code
    """
    old_df_name = step['df_names'][sheet_index]
    new_df_name = step['df_names'][len(step['dfs']) - 1]

    return [f'{new_df_name} = {old_df_name}.copy(deep=True)']


def describe_dataframe_duplicate(
        sheet_index,
        df_names=None
    ):

    if df_names is not None:
        old_df_name = df_names[sheet_index]
        new_df_name = df_names[len(df_names) - 1]
        return f'Duplicated {old_df_name} to {new_df_name}'
    return f'Duplicated a df'


DATAFRAME_DUPLICATE_STEP = {
    'step_version': 1,
    'step_display_name': DATAFRAME_DUPLICATE_DISPLAY_NAME,
    'event_type': DATAFRAME_DUPLICATE_EVENT,
    'step_type': DATAFRAME_DUPLICATE_STEP_TYPE,
    'params': DATAFRAME_DUPLICATE_PARAMS,
    'saturate': None,
    'execute': execute_dataframe_duplicate,
    'transpile': transpile_dataframe_duplicate,
    'describe': describe_dataframe_duplicate
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A simple import, which allows you to import dataframes with the given file_names

NOTE: special care needs to be taken when upgrading this step, as it is used (e.g in the 
api in special ways). Not sure exactly how, but be careful + test it thoroughly.
"""
from mitosheet.errors import make_is_directory_error
import os
import pandas as pd
import json
from os.path import basename, normpath
from mitosheet.utils import add_df_to_step, create_new_step, get_header_renames, make_valid_header

SIMPLE_IMPORT_DISPLAY_NAME = 'Imported CSV Files'

SIMPLE_IMPORT_EVENT = 'simple_import_edit'
SIMPLE_IMPORT_STEP_TYPE = 'simple_import'

SIMPLE_IMPORT_PARAMS = [
    'file_names', # list of strings
]

def file_name_to_df_name(file_name: str):
    # First, we strip off any path beyond the final filename
    file_name = basename(normpath(file_name))

    # We abuse the fact that all valid mito headers are almost valid variable names
    # If they start w/ numbers, we add letters
    possible_var_name = make_valid_header(file_name)
    if len(possible_var_name) == 0 or possible_var_name[0].isnumeric():
        return 'df_' + possible_var_name
    return possible_var_name

def get_dataframe_names(file_names, existing_df_names):
    """
    Helper function for taking a list of file names and turning them into valid
    names for dataframes.

    NOTE:
    1. If there are duplicates, appends onto the end of them to deduplicate.
    2. Avoids overwriting existing df names.
    """
    new_names_inital = [file_name_to_df_name(file_name) for file_name in file_names]

    final_names = []

    # Keep appending names till we get one that doesn't overlap
    for name in new_names_inital:
        curr_name = name
        count = 0
        while curr_name in final_names or curr_name in existing_df_names:
            curr_name = f'{name}_{count}'
            count += 1
        
        final_names.append(curr_name)
    
    return final_names


def execute_simple_import(
        curr_step,
        file_names
    ):
    """
    Imports the files with the given file names into the sheet
    """
    # If any of the files are directories, we throw an error to let
    # the user know
    for file_name in file_names:
        if os.path.isdir(file_name):
            raise make_is_directory_error(file_name)

    # Create a new step
    new_step = create_new_step(curr_step, SIMPLE_IMPORT_STEP_TYPE, deep=False)

    column_header_renames = dict()
    non_tutorial_data = False
    for file_name, df_name in zip(file_names, get_dataframe_names(file_names, new_step['df_names'])):

        # TODO: catch this error in a good way
        df = pd.read_csv(file_name)
        renames = get_header_renames(df.keys())
        if len(renames) > 0:
            # Save that we did these renames
            column_header_renames[df_name] = renames
            # Actually perform any renames we need to
            df.rename(columns=renames, inplace=True)

        add_df_to_step(new_step, df, df_name=df_name)        

    # Save the renames that have occured in the step, for transpilation reasons
    new_step['column_header_renames'] = column_header_renames


    return new_step


def transpile_simple_import(
        step,
        file_names
    ):
    """
    Transpiles a sort step to Python code. 
    """

    code = ['import pandas as pd']
    for file_name, df_name in zip(file_names, step['df_names'][len(step['df_names']) - len(file_names):]):

        code.append(
            f'{df_name} = pd.read_csv(\'{file_name}\')'
        )
        # If we had to rename columns, mark that as well
        if df_name in step['column_header_renames']:
            renames = step['column_header_renames'][df_name]
            code.append(
                f'{df_name}.rename(columns={json.dumps(renames)}, inplace=True)'
            )

    return code


def describe_simple_import(
        file_names,
        df_names=None
    ):
    return f'Imported {", ".join(file_names)}'


SIMPLE_IMPORT_STEP = {
    'step_version': 1,
    'step_display_name': SIMPLE_IMPORT_DISPLAY_NAME,
    'event_type': SIMPLE_IMPORT_EVENT,
    'step_type': SIMPLE_IMPORT_STEP_TYPE,
    'params': SIMPLE_IMPORT_PARAMS,
    'saturate': None,
    'execute': execute_simple_import,
    'transpile': transpile_simple_import,
    'describe': describe_simple_import
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Utilities to help with type functions
"""

from typing import Tuple, Union
import pandas as pd
import numpy as np

# Because type(1) = int, thus 1 is a 'number' in the Mito type system
MITO_PRIMITIVE_TYPE_MAPPING = {
    'boolean': [bool],
    'timestamp': [pd.Timestamp],
    'timedelta': [pd.Timedelta],
    'number': [int, float],
    'string': [str],
}


BOOLEAN_SERIES = 'boolean_series'
DATETIME_SERIES = 'datetime_series'
TIMEDELTA_SERIES = 'timedelta_series'
NUMBER_SERIES = 'number_series'
STRING_SERIES = 'string_series'

# NOTE: these are the dtype strings of series used by pandas!
MITO_SERIES_TYPE_MAPPING = {
    BOOLEAN_SERIES: ['bool'],
    DATETIME_SERIES: ['datetime64[ns]'],
    TIMEDELTA_SERIES: ['timedelta64[ns]'],
    NUMBER_SERIES: ['int64', 'float64'],
    STRING_SERIES: ['str', 'object'],
}

def get_mito_type(obj):

    if isinstance(obj, pd.Series):
        dtype = obj.dtype
        for key, value in MITO_SERIES_TYPE_MAPPING.items():
            if dtype in value:
                return key

    elif isinstance(obj, pd.Timestamp):
        return 'timestamp'
    elif isinstance(obj, pd.Timedelta):
        return 'timedelta'
    else:
        obj_type = type(obj)

        for key, value in MITO_PRIMITIVE_TYPE_MAPPING.items():
            if obj_type in value:
                return key

    return None


def get_nan_indexes(*argv): 
    """
    Given the list of arguments to a function, as *argv, 
    returns a list of row indexes that is True iff one of the series 
    params is NaN at that index. Otherwise the index is False

    This function is called by the filter_nan decorator 
    """

    # we find the max length of the args because we are unsure ahead of time which arg 
    # is a series. we need the max_length to construct the boolean index array
    max_length = -1
    for arg in argv:
        # check the type to make sure an error is not thrown for calling len() on an int
        if isinstance(arg, pd.Series) and len(arg) > max_length:
            max_length = len(arg)

    # if there are no series, then:
    #   1. there are no NaN values
    #   2. all of the args are a length of 1
    if max_length == -1:
        return pd.Series([False], dtype='bool')

    nan_indexes = pd.Series([False for i in range(max_length)], dtype='bool')

    # for each row, check for NaN values in the function
    for arg in argv:
        if isinstance(arg, pd.Series): 
            nan_indexes = nan_indexes | pd.isna(arg)

    return nan_indexes

def put_nan_indexes_back(series, nan_indexes):
    """
    This functions takes a series and a boolean index list that is True 
    if the index in the series should be NaN and false if it should
    be left alone. 

    Returns the series with the NaN values put in

    This function is called by the as_types decorator 
    """
    original_length = len(nan_indexes)
    final_series = []
    real_index = 0
    non_nan_index = 0

    while real_index < original_length:
        if nan_indexes[real_index]:
            final_series.append(np.NaN)
        else:
            final_series.append(series[non_nan_index])
            non_nan_index += 1
        real_index +=1

    final_series = pd.Series(final_series)
    return final_series


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
See mito/mitosheet/steps/README.md for more information about 
how to add a step!

NOTE: this new step refactor is a WIP, and will not be finished
until we remove all the manual step casing that occurs throughout the 
codebase. This is an incremental process that will take time!
"""
# NOTE: init step should not be in the main list
from mitosheet.steps.initalize import INITIALIZE_STEP

from mitosheet.steps.pivot import PIVOT_STEP
from mitosheet.steps.filter import FILTER_STEP
from mitosheet.steps.sort import SORT_STEP
from mitosheet.steps.column_steps.reorder_column import REORDER_COLUMN_STEP
from mitosheet.steps.column_steps.add_column import ADD_COLUMN_STEP
from mitosheet.steps.column_steps.set_column_formula import SET_COLUMN_FORMULA_STEP
from mitosheet.steps.merge import MERGE_STEP
from mitosheet.steps.column_steps.delete_column import DELETE_COLUMN_STEP
from mitosheet.steps.column_steps.rename_column import RENAME_COLUMN_STEP
from mitosheet.steps.import_steps.simple_import import SIMPLE_IMPORT_STEP
from mitosheet.steps.import_steps.raw_python_import import RAW_PYTHON_IMPORT_STEP
from mitosheet.steps.dataframe_steps.dataframe_delete import DATAFRAME_DELETE_STEP
from mitosheet.steps.dataframe_steps.dataframe_duplicate import DATAFRAME_DUPLICATE_STEP
from mitosheet.steps.dataframe_steps.dataframe_rename import DATAFRAME_RENAME_STEP

# All steps must be listed in this variable.
STEPS = [
    PIVOT_STEP,
    REORDER_COLUMN_STEP,
    FILTER_STEP,
    SORT_STEP,
    ADD_COLUMN_STEP,
    SET_COLUMN_FORMULA_STEP,
    MERGE_STEP,
    DELETE_COLUMN_STEP,
    RENAME_COLUMN_STEP,
    SIMPLE_IMPORT_STEP,
    RAW_PYTHON_IMPORT_STEP,
    DATAFRAME_DELETE_STEP,
    DATAFRAME_DUPLICATE_STEP,
    DATAFRAME_RENAME_STEP
]

# A helpful mapping for looking up steps based on the incoming events
EVENT_TYPE_TO_STEP = {
    step['event_type']: step for step in STEPS
}

# We also build a useful lookup mapping for the step type to step object
STEP_TYPE_TO_STEP = {
    step['step_type']: step for step in STEPS
}
STEP_TYPE_TO_STEP[INITIALIZE_STEP['step_type']] = INITIALIZE_STEP # save initalize here as well


from mitosheet.sheet_functions.types.utils import get_mito_type
from mitosheet.save_utils import read_analysis
from mitosheet.steps import STEP_TYPE_TO_STEP
import json


def get_saved_analysis_description(send, event, wsc):
    """
    Gets the description of the steps in the saved analysis, and sends
    it back to the frontend

    Sends '' if its unable to read in for any reason (or this analysis does)
    not exist.
    """
    analysis_name = event['analysis_name']
    analysis_steps = read_analysis(analysis_name)
    
    try:
        if analysis_steps is None:
            send({
                'event': 'api_response',
                'id': event['id'],
                'data': ''
            })
            return
        
        analysis_descriptions = []
        for step_number in analysis_steps['steps']:

            step = analysis_steps['steps'][step_number]
            STEP_OBJ = STEP_TYPE_TO_STEP[step['step_type']]

            # Create the step param object
            params = {key: value for key, value in step.items() if key in STEP_OBJ['params']}

            # Generate the step description
            step_description = STEP_OBJ['describe'](
                **params
            )

            analysis_descriptions.append({
                'step_type': step['step_type'],
                'step_description': step_description
            })

        send({
            'event': 'api_response',
            'id': event['id'],
            'data': json.dumps(analysis_descriptions)
        })
        
    except Exception as e:
        # As not being able to get the steps is not a critical failure, 
        # we return empty data if its not possible.
        send({
            'event': 'api_response',
            'id': event['id'],
            'data': ''
        })


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Tests some helper functions of local deployment.
"""

import pytest

from mitosheet.utils import get_app_directory_from_jupyter_lab_path_output


APP_DIR_TESTS = [
    (
        """Application directory:   /Users/nate/saga-vcs/monorepo/mito/venv/share/jupyter/lab
User Settings directory: /Users/nate/.jupyter/lab/user-settings
Workspaces directory: /Users/nate/.jupyter/lab/workspaces""",
        '/Users/nate/saga-vcs/monorepo/mito/venv/share/jupyter/lab'
    ),
    # Mac, with spaces
    (
        """Application directory:   /Users/nate/space dir/venv/share/jupyter/lab
User Settings directory: /Users/nate/.jupyter/lab/user-settings
Workspaces directory: /Users/nate/.jupyter/lab/workspaces""",
        '/Users/nate/space dir/venv/share/jupyter/lab'
    ),
    (
        """Application directory:   c:\\users\\nater\\mitovenv\\share\\jupyter\\lab
User Settings directory: C:\\Users\\nater\\.jupyter\\lab\\user-settings
Workspaces directory: C:\\Users\\nater\\.jupyter\\lab\\workspaces""",
        'c:\\users\\nater\\mitovenv\\share\\jupyter\\lab'
    ),
    (
        """Application directory:   c:\\users\\nater\\space folder\\mitovenv\\share\\jupyter\\lab
User Settings directory: C:\\Users\\nater\\.jupyter\\lab\\user-settings
Workspaces directory: C:\\Users\\nater\\.jupyter\\lab\\workspaces""",
        'c:\\users\\nater\\space folder\\mitovenv\\share\\jupyter\\lab'
    ),
]
@pytest.mark.parametrize('output, path', APP_DIR_TESTS)
def test_get_app_directory(output, path):
    assert get_app_directory_from_jupyter_lab_path_output(output) == path


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
For going to a datetime series.
"""

from typing import Tuple, Union
import pandas as pd
import numpy as np

from mitosheet.sheet_functions.types.utils import get_mito_type

def to_datetime_series(
        unknown_object,
        on_uncastable_arg_element=('default', np.NaN), # Union[Literal['error'], Tuple[Literal['default'], any]]
    ):
    """
    Converts the given object to a datetime series.
    """
    from_type = get_mito_type(unknown_object)

    # If it is not a series, we put it in a series, and get the type again
    if not from_type.endswith('series'):
        unknown_object = pd.Series([unknown_object])
        from_type = get_mito_type(unknown_object)

    if from_type == 'boolean_series':
        return None
    elif from_type == 'datetime_series':
        return unknown_object
    if from_type == 'number_series':
        return None
    elif from_type == 'string_series':
        # TODO: improve this to work element wise!
        return pd.to_datetime(unknown_object)
    else:
        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.

import json
import uuid 
from copy import deepcopy
import pandas as pd
from typing import List


from mitosheet.steps.column_steps.set_column_formula import SET_COLUMN_FORMULA_EVENT, SET_COLUMN_FORMULA_STEP_TYPE
from mitosheet.steps.filter import execute_filter_column
from mitosheet.steps import EVENT_TYPE_TO_STEP, STEP_TYPE_TO_STEP
from mitosheet.updates import UPDATES
from mitosheet.steps.initalize import execute_initialize_step
from mitosheet.preprocessing import PREPROCESS_STEPS
from mitosheet.utils import dfs_to_json
from mitosheet.transpiler.transpile import transpile
from mitosheet.user.user_utils import get_user_field
from mitosheet.data_in_mito import DataTypeInMito, get_data_type_in_mito
from mitosheet.mito_analytics import log

class WidgetStateContainer():
    """
    Holds the state of the steps, which represents operations that
    have been performed on the input dataframes. 
    """

    def __init__(self, args):
        """
        When initalizing the Widget State Container, we also do preprocessing
        of the arguments that were passed to the mitosheet. 

        All preprocessing can be found in mitosheet/preprocessing, and each of 
        the transformations are applied before the data is considered imported.
        """
        # We just randomly generate analysis names. 
        # We append a UUID to note that this is not an analysis the user has saved.
        self.analysis_name = 'UUID-' + str(uuid.uuid4())

        # The args are a tuple of dataframes or strings, and we start by making them
        # into a list, and making copies of them for safe keeping
        self.original_args = [
            arg.copy(deep=True) if isinstance(arg, pd.DataFrame) else deepcopy(arg) for arg in args
        ]

        # Then, we go through the process of actually preprocessing the args
        for PREPROCESS_STEP in PREPROCESS_STEPS:
            args = PREPROCESS_STEP['execute'](args)

        # Then we initialize the first initalize step
        self.steps = []
        # We display the state that exists after the curr_step_idx is applied,
        # which means you can never see before the initalize step
        self.curr_step_idx = 0

        self.data_type_in_mito: DataTypeInMito = get_data_type_in_mito()

        execute_initialize_step(self, args)

    @property
    def curr_step(self):
        """
        Returns the current step object as a property of the object,
        so reference it with self.curr_step
        """
        return self.steps[self.curr_step_idx]

    @property
    def num_sheets(self):
        """
        Duh. :)
        """
        return len(self.steps[self.curr_step_idx]['dfs'])

    @property
    def dfs(self) -> List[pd.DataFrame]:
        return self.steps[self.curr_step_idx]['dfs']

    @property
    def df_names_json(self):
        """
        A JSON object containing the names of the dataframes
        """
        return json.dumps({'df_names': self.curr_step['df_names']})

    @property
    def sheet_json(self):
        """
        sheet_json contains a serialized representation of the data
        frames that is then fed into the ag-grid in the front-end. 

        NOTE: we only display the _first_ 2,000 rows of the dataframe
        for speed reasons. This results in way less data getting 
        passed around
        """
        return dfs_to_json(self.curr_step['dfs'])
    
    @property
    def df_shape_json(self):
        """
        Returns the df shape (rows, columns) of each dataframe in the 
        current step!
        """
        return json.dumps([
            {'rows': df.shape[0], 'cols': df.shape[1]}
            for df in self.curr_step['dfs']
        ])

    @property
    def column_spreadsheet_code_json(self):
        """
        column_spreadsheet_code_json is a list of all the spreadsheet
        formulas that users have used, for each sheet they have. 
        """
        return json.dumps(self.curr_step['column_spreadsheet_code'])

    @property
    def code_json(self):
        """
        This code json string is sent to the front-end and is what
        ends up getting displayed in the codeblock. 
        """
        return json.dumps(transpile(self))

    @property
    def column_filters_json(self):
        """
        This column_filters list is used by the front end to display
        the filtered icons in the UI
        """
        return json.dumps(self.curr_step['column_filters'])
    
    @property
    def column_type_json(self):
        """
        Returns a list of JSON objects that hold the type of each
        data in each column.
        """
        return json.dumps(self.curr_step['column_type'])

    @property
    def user_email(self):
        """
        Returns the user_email from user.json
        """
        return get_user_field('user_email')

    @property
    def received_tours(self):
        """
        Returns the tours that the user has received
        """
        return json.dumps(get_user_field('received_tours'))

    @property
    def intended_behavior(self):
        """
        Returns the intended_behavior of the user as inputted in the signup flow
        """
        return json.dumps(get_user_field('intended_behavior'))

    @property
    def step_data_list_json(self):
        """
        Returns a list of step data
        """
        step_data_list = []
        for index, step in enumerate(self.steps):
            STEP_OBJ = STEP_TYPE_TO_STEP[step['step_type']]
            params = {key: value for key, value in step.items() if key in STEP_OBJ['params']}
            step_data_list.append({
                'step_id': step['step_id'],
                'step_idx': index,
                'step_type': step['step_type'],
                'step_display_name': STEP_OBJ['step_display_name'],
                'step_description': STEP_OBJ['describe'](
                    df_names=step['df_names'],
                    **params,
                )
            })

        return json.dumps(step_data_list)

    def edit_event_should_overwrite_curr_step(self, edit_event):
        """
        We overwrite the step if the step_ids are shared between the curr step
        and the edit event, or if if it is an set formula event that is setting 
        the formula of a column that was updated the last step
        """
        overwrite_formula_step = self.curr_step['step_type'] == SET_COLUMN_FORMULA_STEP_TYPE \
                and edit_event['type'] == SET_COLUMN_FORMULA_EVENT \
                and self.curr_step['sheet_index'] == edit_event['sheet_index'] \
                and self.curr_step['column_header'] == edit_event['column_header']

        return self.curr_step['step_id'] == edit_event['step_id'] or overwrite_formula_step

    def handle_edit_event(self, edit_event):
        """
        Updates the widget state with a new step that was created
        by the edit_event. Each edit_event creates at most one step. 

        If there is an error in the creation of the new step, this
        function will delete the new, invalid step.
        """
        # NOTE: We ignore any edit if we are in a historical state, for now. This is a result
        # of the fact that we don't allow previous editing currently
        if self.curr_step_idx != len(self.steps) - 1:
            return

        try:
            curr_step = self.steps[-1]

            overwrite = self.edit_event_should_overwrite_curr_step(edit_event)

            if overwrite:
                # If we are overwriting the event, then we set the current
                # step back 1 from the actual current step, so that we start
                # from the correct state
                curr_step = self.steps[-2]             

            step_obj = EVENT_TYPE_TO_STEP[edit_event['type']]

            # Saturate the event
            if step_obj['saturate'] is not None:
                step_obj['saturate'](curr_step, edit_event)

            # Get the params for this event
            params = {key: value for key, value in edit_event.items() if key in step_obj['params']}
            # If it's filter, we need to do a lot of extra work
            # so we sent it into the execute function directly
            if step_obj['step_type'] == 'filter_column':
                execute_filter_column(
                    self,
                    **params
                )
                self.curr_step_idx = len(self.steps) - 1
                # If we made a new step, save the step id of this step
                if 'step_id' not in self.curr_step:
                    self.curr_step['step_id'] = edit_event['step_id']
                return 

            # If the user performs a simple import, update the wsc with 
            # whether they used personal data or not. 
            if step_obj['step_type'] == 'simple_import':
                data_type_in_mito = get_data_type_in_mito(*params['file_names'])

                # If they imported PERSONAL data, set type_of_data_in_mito to personal & log PERSONAL data
                if data_type_in_mito == DataTypeInMito.PERSONAL:
                    log('used_personal_data')
                    self.data_type_in_mito = DataTypeInMito.PERSONAL

                # If they imported TUTORIAL data, and they didn't already import PERSONAL data, set type_of_data_in_mito to TUTORIAL
                if data_type_in_mito == DataTypeInMito.TUTORIAL and not self.data_type_in_mito == DataTypeInMito.PERSONAL:
                    self.data_type_in_mito = DataTypeInMito.TUTORIAL

                # If they imported PROVIDED data, and they didn't already import PERSONAL or TUTORIAL data, 
                # set type_of_data_in_mito to PROVIDED
                if (
                    data_type_in_mito == DataTypeInMito.PROVIDED and 
                    not self.data_type_in_mito == DataTypeInMito.PERSONAL and 
                    not self.data_type_in_mito == DataTypeInMito.TUTORIAL
                ):
                    self.data_type_in_mito = DataTypeInMito.PROVIDED

                            
            # Actually get the new step
            new_step = step_obj['execute'](curr_step, **params)

            # If the new step is None, we dont do anything with it, otherwise we
            # add it to the steps
            if new_step is not None:
                # Save the parameters in the new step
                for key, value in params.items():
                    new_step[key] = value
                
                # Furthermore, we save the step_id
                new_step['step_id'] = edit_event['step_id']

                # Finially, we append this step in the correct location to the steps
                if not overwrite:
                    self.steps.append(new_step)
                else:
                    self.steps[-1] = new_step
            
            # and then update the index of the current step
            self.curr_step_idx = len(self.steps) - 1
        except:

            # We bubble the error up if it occurs
            # https://stackoverflow.com/questions/6593922/letting-an-exception-to-bubble-up
            raise


    def handle_update_event(self, update_event):
        """
        Handles any event that isn't caused by an edit, but instead
        other types of new data coming from the frontend (e.g. the df names 
        or some existing steps).
        """
        for update in UPDATES:
            if update_event['type'] == update['event_type']:
                # Get the params for this event
                params = {key: value for key, value in update_event.items() if key in update['params']}
                # Actually execute this event
                update['execute'](self, **params)
                # And then return
                return

        raise Exception(f'{update_event} is not an update event!')        
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Deletes a dataframe from everywhere in the step.
"""

from mitosheet.utils import create_new_step


DATAFRAME_DELETE_DISPLAY_NAME = 'Deleted a Dataframe'

DATAFRAME_DELETE_EVENT = 'dataframe_delete_edit'
DATAFRAME_DELETE_STEP_TYPE = 'dataframe_delete'

DATAFRAME_DELETE_PARAMS = [
    'sheet_index',
    'old_dataframe_name'
]

def saturate_dataframe_delete(
        curr_step,
        event
    ):
    """
    Adds the old dataframe name to the dataframe_delete event
    """
    sheet_index = event['sheet_index']
    old_dataframe_name = curr_step['df_names'][sheet_index]
    event['old_dataframe_name'] = old_dataframe_name


def execute_dataframe_delete(
        curr_step,
        sheet_index,
        old_dataframe_name
    ):
    """
    Deletes the dataframe as sheet index.
    """
    # Create a new step and save the parameters
    new_step = create_new_step(curr_step, DATAFRAME_DELETE_STEP_TYPE, deep=False)

    # Save the parameters
    new_step['sheet_index'] = sheet_index
    new_step['old_dataframe_name'] = old_dataframe_name

    # Execute the delete
    new_step['column_metatype'].pop(sheet_index)
    new_step['column_type'].pop(sheet_index)
    new_step['column_spreadsheet_code'].pop(sheet_index)
    new_step['added_columns'].pop(sheet_index)
    new_step['column_python_code'].pop(sheet_index)
    new_step['column_evaluation_graph'].pop(sheet_index)
    new_step['column_filters'].pop(sheet_index)
    new_step['dfs'].pop(sheet_index)
    new_step['df_names'].pop(sheet_index)

    return new_step

def transpile_dataframe_delete(
        step,
        sheet_index,
        old_dataframe_name
    ):
    """
    Transpiles a delete step (which doesn't do much, in fairness).
    """
    return [f'del {old_dataframe_name}']


def describe_dataframe_delete(
        sheet_index,
        old_dataframe_name,
        df_names=None
    ):
    return f'Deleted dataframe {old_dataframe_name}'


DATAFRAME_DELETE_STEP = {
    'step_version': 1,
    'step_display_name': DATAFRAME_DELETE_DISPLAY_NAME,
    'event_type': DATAFRAME_DELETE_EVENT,
    'step_type': DATAFRAME_DELETE_STEP_TYPE,
    'params': DATAFRAME_DELETE_PARAMS,
    'saturate': saturate_dataframe_delete,
    'execute': execute_dataframe_delete,
    'transpile': transpile_dataframe_delete,
    'describe': describe_dataframe_delete
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A set_column_formula step, which allows you to set the formula
of a given column in the sheet (and then recalculates this column)
and it's dependents.
"""
from mitosheet.topological_sort import creates_circularity, subgraph_from_starting_column_header, topological_sort_columns
from mitosheet.sheet_functions import FUNCTIONS
from mitosheet.parser import parse_formula
from mitosheet.utils import create_new_step, get_column_filter_type

from mitosheet.errors import (
    make_circular_reference_error, 
    make_no_column_error, 
    make_operator_type_error, 
    make_unsupported_function_error, 
    make_wrong_column_metatype_error
)

SET_COLUMN_FORMULA_DISPLAY_NAME = 'Set Column Formula'

SET_COLUMN_FORMULA_EVENT = 'set_column_formula_edit'
SET_COLUMN_FORMULA_STEP_TYPE = 'set_column_formula'

SET_COLUMN_FORMULA_PARAMS = [
    'sheet_index', # int
    'column_header', # the new column to create
    'old_formula', # the previous formula, or None (if there is None)
    'new_formula', # the new formula for the column
]


def _update_column_formula_in_step(
        step,
        sheet_index,
        column_header,
        old_formula,
        new_formula
    ):
    """
    A  helper function for updating the formula of a column. It assumes
    that the passed information is all _correct_ and will not:
    1. Introduce a circular reference error.
    2. Add an invalid formula.

    It DOES NOT "reexecute" the dataframes, just updates the state variables.
    """

    new_python_code, _, new_dependencies = parse_formula(
        new_formula, 
        column_header
    )

    _, _, old_dependencies = parse_formula(
        old_formula, 
        column_header
    )

    step['column_spreadsheet_code'][sheet_index][column_header] = new_formula
    step['column_python_code'][sheet_index][column_header] = new_python_code

    # Update the column dependency graph
    for old_dependency in old_dependencies:
        step['column_evaluation_graph'][sheet_index][old_dependency].remove(column_header)
    for new_dependency in new_dependencies:
        step['column_evaluation_graph'][sheet_index][new_dependency].add(column_header)


def saturate_set_column_formula_step(
        curr_step,
        event
    ):
    """
    Adds the old formula to the edit event.
    """
    sheet_index = event['sheet_index']
    column_header = event['column_header']
    event['old_formula'] = curr_step['column_spreadsheet_code'][sheet_index][column_header]

def execute_set_column_formula_step(
        curr_step,
        sheet_index,
        column_header,
        old_formula,
        new_formula
    ):
    """
    Sets the column with column_header to have the new_formula, and 
    updates the dataframe as a result.

    Errors if:
    - The given column_header is not a column. 
    - The new_formula introduces a circular reference.
    - The new_formula causes an execution error in any way. 

    In the case of an error, this function rolls back all variables
    variables to their state at the start of this function.
    """
    if column_header not in curr_step['column_metatype'][sheet_index]:
        raise make_no_column_error([column_header])

    # First, we check the column_metatype, and make sure it's a formula
    if curr_step['column_metatype'][sheet_index][column_header] != 'formula':
        raise make_wrong_column_metatype_error(column_header)

    # If nothings changed, there's no work to do
    if (old_formula == new_formula):
        return

    # Then we try and parse the formula
    new_python_code, new_functions, new_dependencies = parse_formula(
        new_formula, 
        column_header
    )
    
    # We check that the formula doesn't reference any columns that don't exist
    missing_columns = new_dependencies.difference(curr_step['column_metatype'][sheet_index].keys())
    if any(missing_columns):
        raise make_no_column_error(missing_columns)

    # The formula can only reference known formulas
    missing_functions = new_functions.difference(set(FUNCTIONS.keys()))
    if any(missing_functions):
        raise make_unsupported_function_error(missing_functions)

    # Then, we get the list of old column dependencies and new dependencies
    # so that we can update the graph
    old_python_code, old_functions, old_dependencies = parse_formula(old_formula, column_header)

    # Before changing any variables, we make sure this edit didn't
    # introduct any circularity
    circularity = creates_circularity(
        curr_step['column_evaluation_graph'][sheet_index], 
        column_header,
        old_dependencies,
        new_dependencies
    )
    if circularity:
        raise make_circular_reference_error()

    # We check out a new step
    new_step = create_new_step(curr_step, SET_COLUMN_FORMULA_STEP_TYPE)

    # Update the column formula, and then execute the new formula graph
    _update_column_formula_in_step(new_step, sheet_index, column_header, old_formula, new_formula)
    _execute(new_step, new_step['dfs'][sheet_index], sheet_index)

    # Finially, update the type of the filters of this column, for all the filters
    # TODO: fix bug where we have to update downstream filter types, but note that
    # it would just be really nice if we didn't have to store the type here, and could
    # just get it dynamically...
    new_type = get_column_filter_type(new_step['dfs'][sheet_index][column_header])
    new_step['column_type'][sheet_index][column_header] = new_type
    new_step['column_filters'][sheet_index][column_header]['filters'] = [
        {'type': new_type, 'condition': filter_['condition'], 'value': filter_['value']} 
        for filter_ in new_step['column_filters'][sheet_index][column_header]['filters']
    ]

    return new_step

def get_details_from_operator_type_error(error):
    """
    We detect operator errors by checking the error string, which has the format:

    If you write 1 + 'true'
    - unsupported operand type(s) for /: 'int' and 'str' 

    If you write 1 + pd.Series(['true'])
    - can only concatenate str (not "int") to str

    If you write 10 ^ pd.Series(['true'])
    - Cannot perform 'rxor' with a dtyped [object] array and scalar of type [bool]
    If you write pd.Series(['true']) ^ 10
    - Cannot perform 'xor' with a dtyped [object] array and scalar of type [bool]

    ^ NOTE: the above two errors makes pretty much no sense at all, but it appears
    to be casting the number to a boolean? So we choose to handle this as a number,
    as this is when you would be 

    Returns None if the passed error is not an operator type error. Otherwise, returns
    a triple of (operator, first argument type, second argument type).
    """
    error_message = str(error)
    # Handle case where standard python primitive types are used incorrectly
    if error_message.startswith('unsupported operand type'):
        # Then, we get the specific operator by checking right before the :
        operator = error_message.split(':')[0].split(' ')[-1].strip()
        # And we get the specific types - see above format for how this parsing works!
        arg_one_type = error_message.split(':')[-1].split('\'')[1]
        arg_two_type = error_message.split(':')[-1].split('\'')[-2]
        return (operator, arg_one_type, arg_two_type)
    # Handle case where pandas dataframes are part of the operator, for a concatenate
    if error_message.startswith('can only concatenate str'):
        # We cannot tell the ordered here (e.g. we don't know if a string was concated
        # to an integer, or the other way around) b/c the error doesn't say so, so we just
        # do our best 
        other_type = error_message.split('"')[1]
        return ('+', 'str', other_type)
    # Handle a ^
    if error_message.startswith('Cannot perform \'rxor\' with a dtyped') or error_message.startswith('Cannot perform \'xor\' with a dtyped'):
        # This error message might be totally wrong, but there is some weird
        # casting that goes on in the error message that makes it hard to tell. As such,
        # we report the most likely error.
        return ('^', 'number', 'str')


    return None


def _execute(step, df, sheet_index):
    """
    Executes the given state variables for  
    """

    topological_sort = topological_sort_columns(step['column_evaluation_graph'][sheet_index])

    for column_header in topological_sort:
        # Exec the code, where the df is the original dataframe
        # See explination here: https://www.tutorialspoint.com/exec-in-python
        try:
            exec(
                step['column_python_code'][sheet_index][column_header],
                {'df': df}, 
                FUNCTIONS
            )
        except TypeError as e:
            # We catch TypeErrors specificially, so that we can case on operator errors, to 
            # give better error messages
            operator_type_error_details = get_details_from_operator_type_error(e)
            if operator_type_error_details is not None:
                # If there is an operator error, we handle it specially, to give the user
                # more information about how to recover
                raise make_operator_type_error(*operator_type_error_details)
            else:
                # If it's not an operator error, we just propagate the error up
                raise e



def transpile_set_column_formula_step(
        step,
        sheet_index,
        column_header,
        old_formula, # NOTE: the change_column_type formula relies on this not being used in the transpile step
        new_formula # same as above
    ):
    """
    Transpiles an set_column_formula step to python code!
    """
    code = []

    # We only look at the sheet that was changed, and sort the columns, taking only
    # those downstream from the changed columns
    subgraph = subgraph_from_starting_column_header(step['column_evaluation_graph'][sheet_index], column_header)
    topological_sort = topological_sort_columns(subgraph)

    # We compile all of their formulas
    for column in topological_sort:
        column_formula_changes = step['column_python_code'][sheet_index][column]
        if column_formula_changes != '':
            # We replace the data frame in the code with it's parameter name!
            column_formula_changes = column_formula_changes.strip().replace('df', f'{step["df_names"][sheet_index]}')
            code.append(column_formula_changes)

    return code


def describe_set_column_formula_step(
        sheet_index,
        column_header,
        old_formula, # NOTE: the change_column_type formula relies on this not being used in the transpile step
        new_formula, # same as above
        df_names=None
    ):

    if df_names is not None:
        df_name = df_names[sheet_index]
        return f'Set {column_header} in {df_name} to {new_formula}'
    return f'Set {column_header} to {new_formula}'


SET_COLUMN_FORMULA_STEP = {
    'step_version': 1,
    'step_display_name': SET_COLUMN_FORMULA_DISPLAY_NAME,
    'event_type': SET_COLUMN_FORMULA_EVENT,
    'step_type': SET_COLUMN_FORMULA_STEP_TYPE,
    'params': SET_COLUMN_FORMULA_PARAMS,
    'saturate': saturate_set_column_formula_step,
    'execute': execute_set_column_formula_step,
    'transpile': transpile_set_column_formula_step,
    'describe': describe_set_column_formula_step
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains an analysis with a single raw import step that imports
a single step into the sheet.
"""

MITO_SIMPLE_RAW_IMPORT_SAVE = {
    'name': 'mito_simple_raw_import',
    'saved_analysis': {
        "version": "0.1.96", 
        "steps": {
            "1": {
                "step_version": 1, 
                "step_type": "raw_python_import", 
                "python_code": "import pandas as pd\ndf = pd.DataFrame({'A': [123]})",
                "new_df_names": ["df"]
            }
        }
    }
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A pivot step, which allows you to pivot data
from an existing dataframe along some keys, and then
aggregate other columns with specific functions.
"""
from mitosheet.utils import add_df_to_step, create_new_step, does_sheet_index_exist_within_step
import pandas as pd
from pandas.core.base import DataError
import pandas as pd

from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
    make_invalid_aggregation_error,
    make_duplicated_pivot_column_error
)

PIVOT_DISPLAY_NAME = 'Pivoted a Dataframe'

PIVOT_EVENT = 'pivot_edit'
PIVOT_STEP_TYPE = 'pivot'

PIVOT_PARAMS = [
    'sheet_index', # int
    'pivot_rows', # list of column_headers, could be empty
    'pivot_columns', # list of column_headers, could be empty
    'values', # a dict from column_header -> aggregation function
]

def saturate_pivot_step(curr_step, event):
    """
    Saturates the pivot table with just a `created_dataframe` key, which
    is useful for logging
    """
    # Case 1 - we have at least one row and at least one value
    created_case_1 = len(event['pivot_rows']) > 0 and len(event['values']) > 0
    # Case 2 - we have at least one column and at least one value
    created_case_2 = len(event['pivot_columns']) > 0 and len(event['values']) > 0
    # TODO: aaron, these are the only two cases, right?
    event['created_non_empty_dataframe'] = created_case_1 or created_case_2

def execute_pivot_step(
        curr_step,
        sheet_index,
        pivot_rows,
        pivot_columns,
        values
    ):
    """
    The function responsible for updating the widget state container
    with a new pivot step.

    If it fails part of the way through, deletes the new pivot step entirely.
    """

    # if the sheets don't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index):
        raise make_no_sheet_error(sheet_index)

    # We check that the pivot by doesn't use any columns that don't exist
    columns_used = set(pivot_rows).union(set(pivot_columns))
    missing_pivot_keys = columns_used.difference(curr_step['column_metatype'][sheet_index].keys())
    if len(missing_pivot_keys) > 0:
        raise make_no_column_error(missing_pivot_keys)

    missing_value_keys = set(values.keys()).difference(curr_step['column_metatype'][sheet_index].keys())
    if len(missing_value_keys) > 0:
        raise make_no_column_error(missing_value_keys)

    # Create a new step
    new_step = create_new_step(curr_step, PIVOT_STEP_TYPE, deep=False)

    # We change the values object here, to make sure that the aggregation functions we use are the ones
    # we want

    # Actually execute the pivoting
    try:
        new_df = _execute_pivot(
            new_step['dfs'][sheet_index], 
            pivot_rows,
            pivot_columns,
            values
        )
    except DataError as e:
        # A data-error occurs when you try to aggregate on a column where the function
        # cannot be applied (e.g. 'mean' on a column of strings)
        print(e)
        # Generate an error informing the user
        raise make_invalid_aggregation_error()
    except ValueError as e:
        # A ValueError occurs when you try to aggregate on a column that is also part of the keys of the pivot table. 
        print(e)
        column_incorrectly_used = str(e).split("\'")[1]
        raise make_duplicated_pivot_column_error(column_incorrectly_used)

    # Add it to the dataframe
    add_df_to_step(new_step, new_df)

    return new_step

def values_to_functions(values):
    """
    Helper function for turning the values mapping sent by the frontend to the values
    mapping to functions that can actually be passed to the pandas pivot function
    """
    new_values = dict()

    for column_header, agg_func_name in values.items():
        # NOTE: this needs to match the values sent from the frontend
        if agg_func_name == 'count unique':
            agg_func = pd.Series.nunique
        else:
            agg_func = agg_func_name
            
        new_values[column_header] = agg_func
    
    return new_values

def _execute_pivot(df, pivot_rows, pivot_columns, values):
    """
    Helper function for executing the pivot on a specific dataframe
    and then aggregating the values with the passed values mapping
    """

    # First, we handle a special case where the only row is also the value
    # the user wants to count. In this case, we
    # This is something that C.M. wants + needs :) 
    if len(pivot_rows) == 1 and len(pivot_columns) == 0 and len(values) == 1 and values == {pivot_rows[0]: 'count'}:
        groupby_obj = df.groupby([pivot_rows[0]], as_index=False)
        return groupby_obj.size()

    # Second, we handle the special case where there are no keys to aggregate on
    # We return an empty dataframe
    if len(pivot_rows) == 0 and len(pivot_columns) == 0 or len(values) == 0:
        return pd.DataFrame(data={})

    values_keys = list(values.keys())

    # Built the args, leaving out any unused values
    args = {}

    if len(pivot_rows) > 0:
        args['index'] = pivot_rows

    if len(pivot_columns) > 0:
        args['columns'] = pivot_columns

    if len(values) > 0:
        args['values'] = values_keys
        args['aggfunc'] = values_to_functions(values)

    pivot_table = df.pivot_table(**args) # type: pd.DataFrame

    # Flatten the column headers
    pivot_table.columns = [
        '_'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col
        for col in pivot_table.columns.values
    ]

    # flatten the column headers & reset the indexes
    pivot_table = pivot_table.rename_axis(None, axis=1).reset_index()

    return pivot_table

def values_to_functions_code(values):
    """
    Helper function for turning the values mapping sent by the frontend to the values
    mapping that works in generated code. Namely, needs to replay Count Unique with the
    function pd.Series.nunique.
    """
    string_values = f'{values}'
    # NOTE: this needs to match the values sent from the frontend
    # also note that we overwrite the quotes around Count Unique
    return string_values.replace('\'count unique\'', 'pd.Series.nunique')

# Helpful constants for code formatting
TAB = '    '
NEWLINE_TAB = f'\n{TAB}'

def build_args_code(
        pivot_rows,
        pivot_columns,
        values
    ):
    """
    Helper function for building an arg string, while leaving
    out empty arguments. 
    """
    values_keys = list(values.keys())

    args = []
    if len(pivot_rows) > 0:
        args.append(f'index={pivot_rows},')

    if len(pivot_columns) > 0:
        args.append(f'columns={pivot_columns},')

    if len(values) > 0:
        args.append(f'values={values_keys},')
        args.append(f'aggfunc={values_to_functions_code(values)}')
        
    return NEWLINE_TAB.join(args)


def transpile_pivot_step(
        step,
        sheet_index,
        pivot_rows,
        pivot_columns,
        values
    ):
    """
    Transpiles a pivot step to python code!
    """
    new_df_name = f'df{len(step["dfs"])}'


    # First, we handle a special case where the only row is also the value the user wants to count. 
    # In this case, we simply count it! This handles a common case users bump into
    if len(pivot_rows) == 1 and len(pivot_columns) == 0 and len(values) == 1 and values == {pivot_rows[0]: 'count'}:
        return [
            f'groupby_obj = {step["df_names"][sheet_index]}.groupby([\'{pivot_rows[0]}\'], as_index=False)',
            f'{new_df_name} = groupby_obj.size()'
        ]

    
    # Second, we handle the special case where there are no keys or values to aggregate on
    # We return an empty dataframe. 
    if len(pivot_rows) == 0 and len(pivot_columns) == 0 or len(values) == 0:
        return [f'{new_df_name} = pd.DataFrame(data={{}})']

    transpiled_code = []
    
    # Pivot 
    pivot_comment = '# Pivot the data'
    pivot_table_args = build_args_code(pivot_rows, pivot_columns, values)
    pivot_table_call = f'pivot_table = {step["df_names"][sheet_index]}.pivot_table({NEWLINE_TAB}{pivot_table_args}\n)'
    transpiled_code.append(f'{pivot_comment}\n{pivot_table_call}')

    # Flatten column headers. 
    # NOTE: this only needs to happen if there is more than one pivot_column, as the columns
    # will be flat if there are no columns pivoted
    if len(pivot_columns) > 0:
        flatten_comment = '# Flatten the column headers'
        flatten_code = f'pivot_table.columns = [{NEWLINE_TAB}\'_\'.join([str(c) for c in col]).strip() if isinstance(col, tuple) else col{NEWLINE_TAB}for col in pivot_table.columns.values\n]'
        transpiled_code.append(f'{flatten_comment}\n{flatten_code}')

    # Finially, reset the column name, and the indexes!
    reset_index_comment = f'# Reset the column name and the indexes'
    reset_index_code = f'{new_df_name} = pivot_table.rename_axis(None, axis=1).reset_index()'
    transpiled_code.append(f'{reset_index_comment}\n{reset_index_code}')

    return transpiled_code

def describe_pivot(
        sheet_index,
        pivot_rows,
        pivot_columns,
        values,
        df_names=None
    ):

    if df_names is not None:
        new_df_name = f'df{len(df_names)}'
        old_df_name = df_names[sheet_index]
        return f'Pivoted {old_df_name} into {new_df_name}'
    return f'Pivoted dataframe {sheet_index}'


PIVOT_STEP = {
    'step_version': 2,
    'step_display_name': PIVOT_DISPLAY_NAME,
    'event_type': PIVOT_EVENT,
    'step_type': PIVOT_STEP_TYPE,
    'params': PIVOT_PARAMS,
    'saturate': saturate_pivot_step,
    'execute': execute_pivot_step,
    'transpile': transpile_pivot_step,
    'describe': describe_pivot
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Contains tests for edit events.
"""
import os
from mitosheet.transpiler.transpile import transpile
import pytest
import pandas as pd
import json
import random

from mitosheet.example import sheet
from mitosheet.tests.test_utils import create_mito_wrapper, create_mito_wrapper_dfs

from mitosheet.save_utils import SAVED_ANALYSIS_FOLDER, write_analysis, read_analysis, read_saved_analysis_names, write_saved_analysis
from mitosheet.errors import EditError
from _pytest._code import ExceptionInfo

# We assume only column A exists
PERSIST_ANALYSIS_TESTS = [
    (0, '=0'),
    (1, '=1'),
    (2, '=A + 1'),
    ('APPLE', '=UPPER(\'apple\')'),
    (2, '=LEFT((A + 1) * 100)'),
    ('APPLE', '=UPPER(LOWER(UPPER(\'apple\')))')
]
@pytest.mark.parametrize("b_value,b_formula", PERSIST_ANALYSIS_TESTS)
def test_recover_analysis(b_value, b_formula):
    mito = create_mito_wrapper([1])
    mito.set_formula(b_formula, 0, 'B', add_column=True)
    # We first write out the analysis
    analysis_name = mito.mito_widget.analysis_name
    write_analysis(mito.mito_widget.widget_state_container)

    df = pd.DataFrame(data={'A': [1]})
    new_mito = create_mito_wrapper_dfs(df)
    new_mito.replay_analysis(analysis_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step

    assert curr_step['column_metatype'][0]['B'] == 'formula'
    assert curr_step['column_spreadsheet_code'][0]['B'] == b_formula
    assert new_mito.mito_widget.widget_state_container.dfs[0]['B'].tolist() == [b_value]
    assert new_mito.mito_widget.column_spreadsheet_code_json == json.dumps(curr_step['column_spreadsheet_code'])


# We assume only column A exists
PERSIST_ANALYSIS_TESTS = [
    (0, '=0'),
    (1, '=1'),
    (2, '=A + 1'),
    ('APPLE', '=UPPER(\'apple\')'),
    (2, '=LEFT((A + 1) * 100)'),
    ('APPLE', '=UPPER(LOWER(UPPER(\'apple\')))')
]
@pytest.mark.parametrize("b_value,b_formula", PERSIST_ANALYSIS_TESTS)
def test_persist_analysis_multi_sheet(b_value, b_formula):
    mito = create_mito_wrapper([1], sheet_two_A_data=[1])
    mito.set_formula(b_formula, 0, 'B', add_column=True)
    mito.set_formula(b_formula, 1, 'B', add_column=True)
    # We first write out the analysis
    analysis_name = mito.mito_widget.analysis_name
    write_analysis(mito.mito_widget.widget_state_container)

    df1 = pd.DataFrame(data={'A': [1]})
    df2 = pd.DataFrame(data={'A': [1]})

    new_mito = create_mito_wrapper_dfs(df1, df2)
    new_mito.replay_analysis(analysis_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step

    assert curr_step['column_metatype'][0]['B'] == 'formula'
    assert curr_step['column_spreadsheet_code'][0]['B'] == b_formula
    assert new_mito.mito_widget.widget_state_container.dfs[0]['B'].tolist() == [b_value]

    assert curr_step['column_metatype'][1]['B'] == 'formula'
    assert curr_step['column_spreadsheet_code'][1]['B'] == b_formula
    assert new_mito.mito_widget.widget_state_container.dfs[1]['B'].tolist() == [b_value]
    
    assert new_mito.mito_widget.column_spreadsheet_code_json == json.dumps(curr_step['column_spreadsheet_code'])
    assert new_mito.mito_widget.code_json == mito.mito_widget.code_json


def test_persist_rename_column():
    mito = create_mito_wrapper([1])
    mito.rename_column(0, 'A', 'NEW_COLUMN')

    analysis_name = mito.mito_widget.analysis_name
    write_analysis(mito.mito_widget.widget_state_container)

    df1 = pd.DataFrame(data={'A': [1]})

    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(analysis_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step

    assert curr_step['dfs'][0].equals(pd.DataFrame(data={'NEW_COLUMN': [1]}))

def test_persisit_delete_column():
    mito = create_mito_wrapper([1])
    mito.delete_column(0, 'A')

    analysis_name = mito.mito_widget.analysis_name
    write_analysis(mito.mito_widget.widget_state_container)

    df1 = pd.DataFrame(data={'A': [1]})

    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(analysis_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step

    assert len(curr_step['dfs'][0].keys()) == 0


def test_save_analysis_update():
    mito = create_mito_wrapper([1])
    mito.add_column(0, 'B')
    mito.delete_column(0, 'A')

    random_name = 'UUID-test_save' + str(random.random())

    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    df1 = pd.DataFrame(data={'A': [1]})

    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(random_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step
    assert curr_step['dfs'][0].keys() == ['B']

def test_save_analysis_update_and_overwrite():
    mito = create_mito_wrapper([1])
    mito.add_column(0, 'B')

    random_name = 'UUID-test_save' + str(random.random())

    # Save it once
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    mito.delete_column(0, 'A')
    mito.delete_column(0, 'B')

    # Save it again
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    df1 = pd.DataFrame(data={'A': [1]})

    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(random_name)

    curr_step = new_mito.mito_widget.widget_state_container.curr_step
    assert len(curr_step['dfs'][0].keys()) == 0

def test_saved_analysis_in_saved_analysis():
    mito = create_mito_wrapper([1])
    mito.add_column(0, 'B')

    random_name = 'test_save' + str(random.random())

    # Save it once
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    saved_analysis_names = read_saved_analysis_names()
    assert random_name in saved_analysis_names

def test_failed_replay_does_not_add_steps():
    # Make an analysis and save it
    mito = create_mito_wrapper([1])
    mito.set_formula('=A + 1', 0, 'B', add_column=True)
    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    # Try and rerun it on a dataframe it cannot be rerun on
    df = pd.DataFrame(data={'A': [1], 'B': [3]})
    new_mito = create_mito_wrapper_dfs(df)

    new_mito.replay_analysis(random_name)

    # Make sure no step was added
    assert len(new_mito.mito_widget.widget_state_container.steps) == 1



def test_pivot_by_replays():
    # Make an analysis and save it
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.pivot_sheet(
        0, 
        ['Name'],
        [],
        {'Height': 'sum'}
    )
    
    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    # Try and rerun it on a dataframe it cannot be rerun on
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Nate'], 'Height': [4, 5]})
    new_mito = create_mito_wrapper_dfs(df1)

    new_mito.replay_analysis(random_name)

    # Make sure no step was added
    wsc = new_mito.mito_widget.widget_state_container
    assert len(wsc.steps) == 2
    assert wsc.steps[1]['step_type'] == 'pivot'
    assert len(wsc.curr_step['dfs']) == 2
    assert wsc.curr_step['dfs'][1].equals(
        pd.DataFrame(data={'Name': ['Nate'], 'Height': [9]})
    )


def test_merge_replays():
    # Make an analysis and save it
    df1 = pd.DataFrame(data={'Name': ['Nate', 'Jake'], 'Height': [4, 5]})
    mito = create_mito_wrapper_dfs(df1)
    mito.merge_sheets(
        0, 
        'Name',
        ['Name', 'Height'],
        0,
        'Name',
        ['Name', 'Height']
    )
    
    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    # Try and rerun it on a dataframe it cannot be rerun on
    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(random_name)

    # Make sure no step was added
    wsc = new_mito.mito_widget.widget_state_container
    assert len(wsc.steps) == 2
    assert wsc.steps[1]['step_type'] == 'merge'
    assert len(wsc.curr_step['dfs']) == 2
    assert wsc.curr_step['dfs'][1].equals(
        pd.DataFrame(data={'Name': ['Nate', 'Jake'], 'Height_df1': [4, 5], 'Height_df1_2': [4, 5]})
    )

TEST_FILE_PATH = 'test_file.csv'

def test_import_replays():
    df = pd.DataFrame(data={'A': [1, 2, 3], 'B': [2, 3, 4]})
    df.to_csv(TEST_FILE_PATH, index=False)

    # Create with no dataframes
    mito = create_mito_wrapper_dfs()
    # And then import just a test file
    mito.simple_import([TEST_FILE_PATH])
    
    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    # Try and rerun it on a dataframe it cannot be rerun on
    new_mito = create_mito_wrapper_dfs()
    new_mito.replay_analysis(random_name)

    os.remove(TEST_FILE_PATH)

    # Make sure no step was added
    wsc = new_mito.mito_widget.widget_state_container
    assert len(wsc.steps) == 2
    assert wsc.steps[1]['step_type'] == 'simple_import'
    assert len(wsc.curr_step['dfs']) == 1
    assert wsc.curr_step['dfs'][0].equals(
        df
    )


def test_replay_analysis_does_not_make_removed_columns():
    df1 = pd.DataFrame(data={'A': [123], 'B': [1234]})
    mito = create_mito_wrapper_dfs(df1)

    mito.add_column(0, 'C')

    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    # Try and rerun it on a dataframe with no column B, and it shouldn't recreate B
    df1 = pd.DataFrame(data={'A': [123]})
    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(random_name)


    assert list(new_mito.mito_widget.widget_state_container.dfs[0].keys()) == ['A', 'C']


def test_upgrades_old_analysis_before_replaying_it():
    write_saved_analysis(
        f'{SAVED_ANALYSIS_FOLDER}/UUID-test-upgrade.json',
        {"1": {"step_version": 1, "step_type": "group", "sheet_index": 0, "group_rows": ["A"], "group_columns": [], "values": {"B": "sum"}}},
        version='0.1.60'
    )

    df = pd.DataFrame({'A': [123], 'B': [123]})
    new_mito = create_mito_wrapper_dfs(df)
    new_mito.replay_analysis('UUID-test-upgrade')

    # This pivot happens to be an identity!
    assert new_mito.mito_widget.widget_state_container.dfs[1].equals(
        df
    )

def test_replay_analysis_can_overwrite_entire_analysis():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    df1 = pd.DataFrame(data={'A': [123]})
    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.rename_column(0, 'A', 'C')
    new_mito.replay_analysis(random_name, clear_existing_analysis=True)
    
    curr_step = new_mito.mito_widget.widget_state_container.curr_step
    assert len(new_mito.mito_widget.widget_state_container.steps) == 2
    assert curr_step['dfs'][0].equals(
        pd.DataFrame({'B': [123]})
    )

def test_rename_analysis():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name = 'UUID-test_save' + str(random.random())
    new_randome_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'rename_analysis_update',
        'old_analysis_name': random_name,
        'new_analysis_name': new_randome_name
    })

    df1 = pd.DataFrame(data={'A': [123]})
    new_mito = create_mito_wrapper_dfs(df1)
    new_mito.replay_analysis(new_randome_name, clear_existing_analysis=True)
    
    curr_step = new_mito.mito_widget.widget_state_container.curr_step
    assert len(new_mito.mito_widget.widget_state_container.steps) == 2
    assert curr_step['dfs'][0].equals(
        pd.DataFrame({'B': [123]})
    )


def test_rename_analysis_with_non_existant_analysis():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name
    })

    new_random_name_one = 'UUID-test_save' + str(random.random())
    new_random_name_two = 'UUID-test_save' + str(random.random())

    with pytest.raises(Exception) as e:
        # We call the event handler directly, so we can catch the error
        mito.mito_widget.widget_state_container.handle_update_event({
            'event': 'update_event',
            'type': 'rename_analysis_update',
            'old_analysis_name': new_random_name_one,
            'new_analysis_name': new_random_name_two
        })
    assert(str(e.value)) == f'Invalid rename, with old and new analysis are {new_random_name_one} and {new_random_name_two}'


def test_rename_analysis_with_already_existing_analysis():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name_one = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name_one
    })

    random_name_two = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name_two
    })

    with pytest.raises(Exception) as e:
        # We call the event handler directly, so we can catch the error
        mito.mito_widget.widget_state_container.handle_update_event({
            'event': 'update_event',
            'type': 'rename_analysis_update',
            'old_analysis_name': random_name_one,
            'new_analysis_name': random_name_two
        })
    
    assert(str(e.value)) == f'Invalid rename, with old and new analysis are {random_name_one} and {random_name_two}'


def test_delete_analysis_deletes():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name_one = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name_one
    })

    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'delete_analysis_update',
        'analysis_name': random_name_one
    })

    assert read_analysis(random_name_one) is None


def test_delete_analysis_fails_on_nonexistant_analysis():
    df = pd.DataFrame({'A': [123]})
    mito = create_mito_wrapper_dfs(df)
    mito.rename_column(0, 'A', 'B')

    random_name_one = 'UUID-test_save' + str(random.random())
    mito.mito_widget.receive_message(mito.mito_widget, {
        'event': 'update_event',
        'type': 'save_analysis_update',
        'analysis_name': random_name_one
    })

    with pytest.raises(Exception) as e:
        # We call the event handler directly, so we can catch the error
        mito.mito_widget.widget_state_container.handle_update_event({
            'event': 'update_event',
            'type': 'delete_analysis_update',
            'analysis_name': random_name_one + "1",
        })
        
    assert str(e.value) == f'Cannot delete {random_name_one + "1"} as it does not exist'



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
Exports the transpile function, which takes the backend widget
container and generates transpiled Python code.
"""

from mitosheet.preprocessing import PREPROCESS_STEPS
from mitosheet.steps import STEPS

from mitosheet.steps.column_steps.add_column import ADD_COLUMN_STEP_TYPE
from mitosheet.steps.column_steps.set_column_formula import SET_COLUMN_FORMULA_STEP_TYPE


IN_PREVIOUS_STEP_COMMENT = '# You\'re viewing a previous step. Click fast forward in the Mitosheet above to see the full analysis.'

def transpile(widget_state_container):
    """
    Takes the Python code in the widget_state_container and linearizes it
    so it can be consumed by the front-end. 
    
    When there are multiple sheets, the first sheets code is first, followed
    by the second sheets code, etc. 
    """

    code = []
    filled_steps = 1

    # First, we transpile all the preprocessing steps
    for PREPROCESS_STEP in PREPROCESS_STEPS:
        preprocess_code = PREPROCESS_STEP['transpile'](widget_state_container)
        if len(preprocess_code) > 0:
            code.extend(preprocess_code)

    # We only transpile up to the currently checked out step
    for step_idx, step in enumerate(widget_state_container.steps[:widget_state_container.curr_step_idx + 1]):

        step_code = [f'# Step {filled_steps}']

        for new_step in STEPS:
            if step['step_type'] == new_step['step_type']:
                # Get the params for this event
                params = {key: value for key, value in step.items() if key in new_step['params']}
                # Actually execute this event
                step_code.extend(new_step['transpile'](
                    step,
                    **params
                ))

        # If we don't write any code, skip
        if len(step_code) == 1:
            continue

        filled_steps += 1
        code.extend(step_code)

    # If we have a historical step checked out, then we add a comment letting
    # the user know this is the case
    if widget_state_container.curr_step_idx != len(widget_state_container.steps) - 1:
        code.append(IN_PREVIOUS_STEP_COMMENT)

    return {
        'imports': f'from mitosheet import *',
        'code': code
    }


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

"""
A merge step, which allows you to merge two dataframes
together.
"""
from mitosheet.utils import add_df_to_step, create_new_step, does_sheet_index_exist_within_step
from mitosheet.errors import (
    make_no_sheet_error,
    make_no_column_error,
)

MERGE_DISPLAY_NAME = 'Merged Two Dataframes'

MERGE_EVENT = 'merge_edit'
MERGE_STEP_TYPE = 'merge'

MERGE_PARAMS = [
    'sheet_index_one', # First df to merge
    'merge_key_one', # Key to merge on
    'selected_columns_one', # The columns to keep from the first sheet 
    'sheet_index_two', # Second df to merge
    'merge_key_two', # The second key to merge on 
    'selected_columns_two', # The columns to keep from the second sheet
]

def execute_merge_step(
        curr_step,
        sheet_index_one: int,
        merge_key_one: str,
        selected_columns_one,
        sheet_index_two: int,
        merge_key_two: str,
        selected_columns_two
    ):
    """
    Creates a new sheet by merging sheet_index_one and sheet_index_two together
    on the keys merge_key_one and merge_key_two respectively. 
    
    Note, that merge does not treat np.NaN = np.NaN, so NaN keys won't be matched 
    with anything, making any column in the second sheet NaN for that row in the resulting merged sheet.
    
    The merged sheet will contain all of the columns from sheet_index_one 
    and sheet_index_two

    If either merge key does not exist, it raises an exception.
    """
    # if the sheets don't exist, throw an error
    if not does_sheet_index_exist_within_step(curr_step, sheet_index_one):
        raise make_no_sheet_error(sheet_index_one)

    if not does_sheet_index_exist_within_step(curr_step, sheet_index_two):
        raise make_no_sheet_error(sheet_index_two)

    # We check that the merge doesn't use any columns that don't exist
    missing_sheet_one_key = {merge_key_one}.difference(curr_step['column_metatype'][sheet_index_one].keys())
    if any(missing_sheet_one_key):
        raise make_no_column_error(missing_sheet_one_key)

    missing_sheet_two_key = {merge_key_two}.difference(curr_step['column_metatype'][sheet_index_two].keys())
    if any(missing_sheet_two_key):
        raise make_no_column_error(missing_sheet_two_key)

    # If no errors we create a new step for this merge
    new_step = create_new_step(curr_step, MERGE_STEP_TYPE, deep=False)

    new_df = _execute_merge(
        new_step,
        new_step['dfs'], 
        sheet_index_one,
        merge_key_one,
        selected_columns_one,
        sheet_index_two,
        merge_key_two,
        selected_columns_two
    )    

    # Add this dataframe to the current step!
    add_df_to_step(new_step, new_df)

    return new_step

def _execute_merge(
        new_step,
        dfs, 
        sheet_index_one,
        merge_key_one, 
        selected_columns_one,
        sheet_index_two,
        merge_key_two,
        selected_columns_two
    ):
    """
    Executes a merge on the sheets with the given indexes, merging on the 
    given keys, and only keeping the selection columns from each df.
    
    TODO: figure out how to simulate VLOOKUP style work better here...
    """
    # We drop duplicates to avoid pairwise duplication on the merge.
    temp_df = dfs[sheet_index_two].drop_duplicates(subset=merge_key_two)

    # Then we delete all the columns from each we don't wanna keep
    deleted_columns_one = set(dfs[sheet_index_one].keys()).difference(set(selected_columns_one))
    deleted_columns_two = set(dfs[sheet_index_two].keys()).difference(set(selected_columns_two))

    df_one_cleaned = dfs[sheet_index_one].drop(deleted_columns_one, axis=1)
    df_two_cleaned = temp_df.drop(deleted_columns_two, axis=1)

    # Finially, we perform the merge!
    df_one_name = new_step['df_names'][sheet_index_one]
    df_two_name = new_step['df_names'][sheet_index_two]
    # We make sure the suffixes aren't the same, as otherwise we might end up with 
    # one df with duplicated column headers
    suffix_one = df_one_name
    suffix_two = df_two_name if df_two_name != df_one_name else f'{df_two_name}_2'

    return df_one_cleaned.merge(df_two_cleaned, left_on=[merge_key_one], right_on=[merge_key_two], how='left', suffixes=[f'_{suffix_one}', f'_{suffix_two}'])

def transpile_merge_step(
        step,
        sheet_index_one,
        merge_key_one, 
        selected_columns_one,
        sheet_index_two,
        merge_key_two,
        selected_columns_two
    ):
    """
    Transpiles a merge step to python code!
    """

    # update df indexes to start at 1
    df_one_name = step['df_names'][sheet_index_one]
    df_two_name = step['df_names'][sheet_index_two]
    df_new_name = step['df_names'][len(step['dfs']) - 1]

    # Now, we build the merge code (starting with the code for dropping duplicates)
    merge_code = [
        f'temp_df = {df_two_name}.drop_duplicates(subset=\'{merge_key_two}\')',
    ]

    # If we are only taking some columns, write the code to drop the ones we don't need!
    deleted_columns_one = set(step['dfs'][sheet_index_one].keys()).difference(set(selected_columns_one))
    deleted_columns_two = set(step['dfs'][sheet_index_two].keys()).difference(set(selected_columns_two))
    if len(deleted_columns_one) > 0:
        merge_code.append(
            f'{df_one_name}_tmp = {df_one_name}.drop({list(deleted_columns_one)}, axis=1)'
        )
    if len(deleted_columns_two) > 0:
        merge_code.append(
            f'{df_two_name}_tmp = temp_df.drop({list(deleted_columns_two)}, axis=1)'
        )

    # If we drop columns, we merge the new dataframes
    df_one_to_merge = df_one_name if len(deleted_columns_one) == 0 else f'{df_one_name}_tmp'
    df_two_to_merge = 'temp_df' if len(deleted_columns_two) == 0 else f'{df_two_name}_tmp'

    # We insist column names are unique in dataframes, so while we default the suffixes to be the dataframe
    # names
    suffix_one = df_one_name
    suffix_two = df_two_name if df_two_name != df_one_name else f'{df_two_name}_2'

    # Finially append the merge
    merge_code.append(
        f'{df_new_name} = {df_one_to_merge}.merge({df_two_to_merge}, left_on=[\'{merge_key_one}\'], right_on=[\'{merge_key_two}\'], how=\'left\', suffixes=[\'_{suffix_one}\', \'_{suffix_two}\'])'
    )

    # And then save it
    return merge_code


def describe_merge(
        sheet_index_one,
        merge_key_one, 
        selected_columns_one,
        sheet_index_two,
        merge_key_two,
        selected_columns_two,
        df_names=None
    ):

    if df_names is not None:
        df_one_name = df_names[sheet_index_one]
        df_two_name = df_names[sheet_index_two]
        return f'Merged {df_one_name} and {df_two_name}'
    return f'Merged dataframes {sheet_index_one} and {sheet_index_two}'


MERGE_STEP = {
    'step_version': 1,
    'step_display_name': MERGE_DISPLAY_NAME,
    'event_type': MERGE_EVENT,
    'step_type': MERGE_STEP_TYPE,
    'params': MERGE_PARAMS,
    'saturate': None,
    'execute': execute_merge_step,
    'transpile': transpile_merge_step,
    'describe': describe_merge
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.

import pytest

from ipykernel.comm import Comm
from ipywidgets import Widget
from mitosheet.save_utils import _get_all_analysis_filenames, _delete_analyses

class MockComm(Comm):
    """A mock Comm object.

    Can be used to inspect calls to Comm's open/send/close methods.
    """
    comm_id = 'a-b-c-d'
    kernel = 'Truthy'

    def __init__(self, *args, **kwargs):
        self.log_open = []
        self.log_send = []
        self.log_close = []
        super(MockComm, self).__init__(*args, **kwargs)

    def open(self, *args, **kwargs):
        self.log_open.append((args, kwargs))

    def send(self, *args, **kwargs):
        self.log_send.append((args, kwargs))

    def close(self, *args, **kwargs):
        self.log_close.append((args, kwargs))

_widget_attrs = {}
undefined = object()


@pytest.fixture
def mock_comm():
    _widget_attrs['_comm_default'] = getattr(Widget, '_comm_default', undefined)
    Widget._comm_default = lambda self: MockComm()
    _widget_attrs['_ipython_display_'] = Widget._ipython_display_
    def raise_not_implemented(*args, **kwargs):
        raise NotImplementedError()
    Widget._ipython_display_ = raise_not_implemented

    yield MockComm()

    for attr, value in _widget_attrs.items():
        if value is undefined:
            delattr(Widget, attr)
        else:
            setattr(Widget, attr, value)



@pytest.fixture(scope="session", autouse=True)
def cleanup_files():
    """
    This fixture is responsible for cleaning up all saved analyses that result from
    test runs. 

    It does so by reading in all analyses at the start of the testing session, 
    and then removing any new analyses at the end. 

    NOTE: if you use the app while the tests are running, it might delete
    your analyses - but this is really rare!
    """
    # This test runs before 
    old_analysis_filenames = _get_all_analysis_filenames()

    yield # All tests run here

    # This code runs after all the tests
    curr_analysis_filenames = _get_all_analysis_filenames()
    # Find all the new analysis file names (generated by tests)
    new_analysis_filenames = curr_analysis_filenames.difference(old_analysis_filenames)
    # Delete them
    _delete_analyses(new_analysis_filenames)